'use strict';

const async = require('async');
const caller = require('caller');
const path = require('path');
const jsxl = require('./jsxl');
const filters = new Map();

jsxl.useFilters({
	Plugin: { 
		$type: { // filter plugin factory
			active: { $optional: true, $type: Boolean },
			plugin: { $any: [ Function, Object ] },
			filter: { $optional: true, $type: 'Filter' },
			init:   { $optional: true, $type: Function },
//			exit:   { $optional: true, $type: Function },
		},
		$transform: (context, factory, next) => {
			if (factory.plugin.constructor === Object && 'filter' in factory) return next('.filter is excessive, filter is only necessary when associated plugin is a function')
			next();
		}
	}
});

const extend = (file, input, next) => {
	const root = file[0] == '.' ? caller().match(/(.*)[\/\\]/)[1]||'' : ''; // calculate absolute root of path
	_load(root, file, input, next);
}

module.exports.load = (file, input, filter, next) => {
	const root = file[0] == '.' ? caller().match(/(.*)[\/\\]/)[1]||'' : ''; // calculate absolute root of path
	_load(root, file, input, filter, next)
}

const _load = (root, file, input, filter, next) => {
	if (next === undefined) {
		next = filter;
		filter = undefined;
	}
	let output;
	async.waterfall([
		next => {
			let factory; try {
				factory = require(path.join(root, file)); // load plugin factory
			} catch(err) {
				err.message = `Cannot load plugin factory module '${file}': ${err.message}`;
				return next(err);
			}
			next(null, factory);
		},
		(factory, next) => { 
			let data = {}; data[path.basename(file)] = factory;
			jsxl(data, 'Plugin', next); 
		},
		(factory, next) => {
			if (factory.active === false) {
				console.log(`${file} is inactive, plugin not loaded`);
				return next(true); // skip if inactive
			}
			if (factory.filter) return jsxl.execute({ input }, factory.filter, (err, input) => next(err, factory, input)); // filter plugin input
			next(null, factory, input);
		},
		(factory, input, next) => {
			if (factory.init) return factory.init(err => next(err, factory, input));
			next(null, factory, input);
		},
		(factory, input, next) => {
			if (factory.plugin.constructor === Object) return next(null, factory.plugin); // construct IS plugin 
			factory.plugin(extend, module.exports.load, input, (err, plugin) => next(err, plugin)); // construct plugin from factory 
		},
		(plugin, next) => {
			if (plugin === undefined) {
				console.log(`${file} has inactive parent plugin, plugin not loaded`);
				return next(true); // skip if inactive
			}
			if (filter === undefined) return next(null, filter, plugin);
			let compiled; if (compiled = filters.get(filter)) return next(null, compiled, plugin); // if plugin filter already compiled
			jsxl.compile(filter, (err, compiled) => next(err, compiled, plugin)); // compile plugin filter
		},
		(compiled, plugin, next) => {
			if (compiled === undefined) return next(null, plugin);
			filters.set(filter, compiled); // keep compiled plugin filter 
			let data = {}; data[`${path.basename(file)}.plugin`] = plugin;
			jsxl.execute(data, compiled, next); // filter constructed plugin 
		},
		(plugin, next) => {
			next(null, output = plugin)
		}
	], err => next(err === true ? null : err, err ? undefined : output));
}